from unittest import TestCase

from musicscore.exceptions import TupletNormalTypeError
from musicscore.tuplet import Tuplet
from musicxml import XMLTimeModification


class TestTuplet(TestCase):
    def test_init(self):
        t = Tuplet(actual_notes=3, normal_notes=2, normal_type='eighth', bracket_type='start', bracket_number=1)
        assert t.actual_notes == 3
        assert t.normal_notes == 2
        assert t.normal_type == 'eighth'
        assert t.bracket_type == 'start'
        assert t.bracket_number == 1

    def test_property_exceptions(self):
        t = Tuplet()
        with self.assertRaises(ValueError):
            t.actual_notes = -1
        with self.assertRaises(ValueError):
            t.actual_notes = 0
        with self.assertRaises(ValueError):
            t.actual_notes = 1
        with self.assertRaises(TypeError):
            t.actual_notes = None
        with self.assertRaises(TypeError):
            t.actual_notes = '1'

        with self.assertRaises(ValueError):
            t.normal_notes = -1
        with self.assertRaises(ValueError):
            t.normal_notes = 0
        with self.assertRaises(ValueError):
            t.normal_notes = 1

        with self.assertRaises(TypeError):
            t.normal_notes = '1'

        with self.assertRaises(ValueError):
            t.normal_type = 'bla'

        with self.assertRaises(ValueError):
            t.bracket_type = 'bla'

        with self.assertRaises(ValueError):
            t.bracket_number = -1
        with self.assertRaises(ValueError):
            t.bracket_number = 0
        with self.assertRaises(TypeError):
            t.bracket_number = '1'

    def test_default_values(self):
        t = Tuplet()
        assert t.actual_notes == 3
        assert t.normal_notes == 2
        assert t.normal_type == 'eighth'
        assert t.bracket_type is None
        assert t.bracket_number is None

    def test_tuplet_get_xml_time_modification(self):
        t = Tuplet()
        tm = t.get_xml_time_modification()
        assert isinstance(tm, XMLTimeModification)
        assert tm.xml_actual_notes.value_ == t.actual_notes
        assert tm.xml_normal_notes.value_ == t.normal_notes
        assert tm.xml_normal_type.value_ == t.normal_type

    def test_tuplet_get_xml_tuplet(self):
        t = Tuplet()
        assert t.get_xml_tuplet() is None
        assert t.bracket_number is None
        t.bracket_type = 'start'
        assert t.bracket_number == 1
        assert t.get_xml_tuplet().bracket == 'yes'
        assert t.get_xml_tuplet().type == t.bracket_type
        assert t.get_xml_tuplet().number == 1
        t.bracket_number = 2
        assert t.bracket_number == 2
        assert t.get_xml_tuplet().number == 2

    def test_tuplet_quarter_duration(self):
        t = Tuplet()
        assert t.quarter_duration == 1
        t.quarter_duration = 1.2
        with self.assertRaises(TupletNormalTypeError):
            t.normal_type

    def test_tuplet_automatic_type(self):
        t = Tuplet(quarter_duration=2)
        assert t.normal_type == 'quarter'
        t.quarter_duration = 1 / 2
        assert t.normal_type == '16th'



from typing import Optional

from musicscore.exceptions import TupletNormalTypeError
from musicxml import XMLTimeModification, XMLTuplet

TUPLETACTUALTONORMALNOTES = {3: 2, 5: 4, 6: 4, 7: 4, 9: 8, 10: 8, 11: 8, 12: 8, 13: 8, 14: 8, 15: 8}
TUPLETNORMALTYPES = {8: '32nd', 4: '16th', 2: 'eighth', 1: 'quarter', 0.5: 'half'}


class Tuplet:
    def __init__(self, actual_notes=3, normal_notes=None, normal_type=None, bracket_type=None, bracket_number=None,
                 quarter_duration=1):
        self._actual_notes = None
        self._normal_notes = None
        self._normal_type = None
        self._bracket_type = None
        self._bracket_number = None
        self._quarter_duration = None
        self.actual_notes = actual_notes
        self.normal_notes = normal_notes
        self.normal_type = normal_type
        self.bracket_type = bracket_type
        self.bracket_number = bracket_number
        self.quarter_duration = quarter_duration

    @property
    def actual_notes(self) -> int:
        return self._actual_notes

    @actual_notes.setter
    def actual_notes(self, val):
        if not isinstance(val, int):
            raise TypeError
        if val < 2:
            raise ValueError
        self._actual_notes = val

    @property
    def normal_notes(self) -> int:
        if self._normal_notes is None:
            return TUPLETACTUALTONORMALNOTES.get(self.actual_notes)
        return self._normal_notes

    @normal_notes.setter
    def normal_notes(self, val):
        if val is not None:
            if not isinstance(val, int):
                raise TypeError
            if val < 2:
                raise ValueError
        self._normal_notes = val

    @property
    def normal_type(self) -> str:
        if self._normal_type is None:
            try:
                return TUPLETNORMALTYPES[self.normal_notes / self.quarter_duration]
            except KeyError:
                raise TupletNormalTypeError(
                    f'Cannot find normal type of normal_notes {self.normal_notes} with quarter_duration {self.quarter_duration}. You can set normal_type manually.')
        return self._normal_type

    @normal_type.setter
    def normal_type(self, val):
        permitted = [None, '1024th', '512th', '256th', '128th', '64th', '32nd', '16th', 'eighth', 'quarter', 'half',
                     'whole', 'breve', 'long', 'maxima']
        if val not in permitted:
            raise ValueError(f'Permitted type values are: {permitted}')
        self._normal_type = val

    @property
    def bracket_type(self) -> Optional[str]:
        return self._bracket_type

    @bracket_type.setter
    def bracket_type(self, val):
        permitted = [None, 'start', 'stop']
        if val not in permitted:
            raise ValueError(f'Permitted bracket_type values are: {permitted}')
        self._bracket_type = val

    @property
    def bracket_number(self) -> int:
        if self._bracket_number is None and self.bracket_type:
            return 1
        return self._bracket_number

    @bracket_number.setter
    def bracket_number(self, val):
        if val is not None:
            if not isinstance(val, int):
                raise TypeError
            if val < 1:
                raise ValueError
        self._bracket_number = val

    @property
    def ratio(self):
        return (self.actual_notes, self.normal_notes)

    def get_xml_time_modification(self) -> 'XMLTimeModification':
        output = XMLTimeModification()
        output.xml_actual_notes = self.actual_notes
        output.xml_normal_notes = self.normal_notes
        output.xml_normal_type = self.normal_type
        return output

    def get_xml_tuplet(self) -> XMLTuplet:
        if self.bracket_type:
            output = XMLTuplet(type=self.bracket_type, number=self.bracket_number)
            if self.bracket_type == 'start':
                output.bracket = 'yes'
            return output



from musicscore import Part, QuarterDuration
from musicscore.chord import GraceChord
from musicscore.exceptions import ChordException
from musicscore.tests.util import IdTestCase
from musicxml.xmlelement.xmlelement import XMLType


class TestGraceChord(IdTestCase):
    def test_grace_chord_init(self):
        gch = GraceChord(60)
        assert gch.quarter_duration == 0
        assert isinstance(gch.quarter_duration, QuarterDuration)
        gch.quarter_duration = 0
        with self.assertRaises(ChordException):
            gch.quarter_duration = 2

    def test_grace_chord_type(self):
        p = Part('p1')
        gch = GraceChord(60)

        gch.type = '16th'
        p.add_chord(gch)
        p.finalize()
        assert p.get_chords()[0].notes[0].xml_type.value_ == '16th'
